/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2003/05/22 12:31:36 $ $Revision: 1.5 $ */
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <zlib.h>
#include <stdlib.h>
#include <unistd.h>
#include "io.h"

/* field is the number of fields read in the line s. bp will
 * point to the position of the first unread field */
int move_bp(char *s,int fields)
{
  int a;

  a=0;
  do {
    /* first read whitespace */
    while ((s[a]!='\0') && (isspace(s[a]))) a++;
    if (!isspace(s[a])) {
      /* then field */
      while ((s[a]!='\0') && (!isspace(s[a]))) a++;
      if (isspace(s[a])) fields--;
    }
  } while ((s[a]!='\0') && (fields>0));
  return(a);
}

/* Reads a field from io.buff and returns it. If field is an empty
 * string moves pointer by one field */
int myscan(char *format,void *value)
{
  int a;

  if (strlen(format)>0) {
    a=sscanf(&io.buff[io.bp],format,value);
    if (a) {
      io.bp+=move_bp(&io.buff[io.bp],a);
    }
  } else {
    io.bp+=move_bp(&io.buff[io.bp],1);
  }
  return(a);
}

/* reads l bytes from the stream */
long myread(FILE *in,long l)
{
  int rv;

  if (l>io.linelen) {
    free(io.buff);
    io.linelen=l*2;
    io.buff=(char *)calloc(io.linelen,sizeof(char));
  }
  if (io.filetype) {
    rv=gzread((gzFile) in,io.buff,(unsigned)l);
  } else {
    rv=fread(io.buff,1,(size_t)l,in);
  }
  io.bp=0;
  return(rv);
}

/* My scanf. reads a line and stores it in io.buff */
int mygetline(FILE *in)
{
  int ok,l;
  char *dummy;

  dummy=io.buff;
  l=io.linelen;
  io.bp=0;
  do {
    ok=0;
    if (io.filetype) {
      gzgets(in,&io.buff[io.bp],l+(io.bp>0));
    } else {
      fgets(&io.buff[io.bp],l+(io.bp>0),in);
    }
    if (((int)strlen(io.buff)==io.linelen-1) 
          && (io.buff[io.linelen-2]!='\n') && (!io.eof(in))) {
      /* line is longer than buffer */
      ok=1;
      l=io.linelen;io.linelen*=2;
      /* copy and double buffer */
      dummy=io.buff;
      io.buff=(char *)calloc(io.linelen,sizeof(char));
      memcpy(io.buff,dummy,l*sizeof(char));
      free(dummy);
      io.bp=l-1;
    }
  } while (ok);
  io.bp=0;
  return(io.eof(in));
}

void io_setfiletype(int val)
{
  io.filetype=val;
  if (!io.filetype) {
    io.open=fopen;
    io.close=fclose;
    io.print=fprintf;
    io.tell=ftell;
    io.seek=fseek;
    io.rewind=rewind;
    io.eof=feof;
  } else {
    io.open=(FILE * (*) (const char *,const char *)) gzopen;
    io.close=(int (*) (FILE *)) gzclose;
    io.print=(int (*) (FILE *,const char *,...)) gzprintf;
    io.tell=(long (*) (FILE *)) gztell;
    io.seek=(int (*) (FILE *,long,int)) gzseek;
    io.rewind=(void (*) (FILE *)) gzrewind;
    io.eof=(int (*) (FILE *)) gzeof;
  }
}

int io_detfiletype()
{
  FILE *in;
  unsigned char c1,c2;

  in=fopen(io.filename,"rb");
  if (in==NULL) return(0);
  fscanf(in,"%c%c",&c1,&c2);
  fclose(in);
  io_setfiletype(((c1==31) && (c2==139)));
  return(1);
}

void io_setfilename(char *s)
{
  char *buff,*rv;
  int l;

  if (io.filename!=NULL) free(io.filename);
  if ((s[0]=='/') || (s[0]=='~')) io.filename=strdup(s);
  else {
    l=1000;
    do {
      buff=(char *) malloc(l);
      rv=getcwd(buff,l);
      if (rv==NULL) {
	free(buff);
	l*=2;
      }
    } while (rv==NULL);
    io.filename=(char *) malloc(strlen(s)+strlen(buff)+3);
    sprintf(io.filename,"%s/%s",buff,s);
    free(buff);
  }
}

void reset_io()
{
  io.linelen=1000;
  io.filename=NULL;
  io.in=NULL;
  io.buff=(char *)calloc(io.linelen,sizeof(char));
  io.getline=mygetline;
  io.scan=myscan;
  io.read=myread;
  io_setfiletype(0);
}
