/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2004/02/10 14:57:32 $ $Revision: 1.30 $ */
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkgl.h>
#include <stdlib.h>
#include <stdio.h>
#include "global.h"
#include "rpn.h"
#include "body.h"

/* prototypes */
void FrameOff(GtkWidget *,gpointer);
int RPN_decode(char *,RPN **,char **,int);
void RPN_Free(RPN **);
void RedrawTex(void);
int ilog(double);
void RedoRPNdecode(GtkWidget *,gpointer);

const char *obj_name[13] = { "EOL", "Wall", "Ball", "Cylinder", "Disk",
			    "BallCC", "Force", "StBall", "DTEdge",
			    "DGBall", "DGEllipsoid", "XYZBall" };

const char cef_relstr[6][2]={ "=", ">", "<", "R", "G", "B" };
const char cef_parmainstr[CEF_PARNUM][10]={ "Num", "Base", "Pos", "Orient",
  		"Coordnum", "Velocity", "Force", "AngVelo", "Troque",
		"TrVelo", "TrForce", "TrAngVelo" };
const char cef_parstr[CEF_PARNUM][CEF_SUBPARNUM][4]={ { "num", "" },
  		{ "dx", "dy", "dz", "i1", "i2" },
  		{ "x", "y", "z", "" }, { "rx", "ry", "rz", "phi", "" },
		{ "Z", "" }, { "vx", "vy", "vz", "" },
		{ "fx", "fy", "fz", "" }, { "omx", "omy", "omz", "" },
		{ "tx", "ty", "tz", "" }, { "tvx", "tvy", "tvz", "" },
		{ "tfx", "tfy", "tfz", "" }, { "tox", "toy", "toz", "" },
		{ "pos", "t", "fpos", "" } /* this should be always the last */
		};
const char cef_strrelstr[8][6]={ "Expr.", "R", "G", "B", "RBG",
				 "Arrow", "Size", "Cut" };

int Move_Cut(double dy)
{
  double d;
  int l;

  if (bg.ac->strrel!=7) return(0);
  d=atof(bg.ac->str);
  free(bg.ac->str);
  d+=dy/gl.f/100.0;
  l=ilog(d);
  bg.ac->str=(char *) calloc(l+10,sizeof(char));
  sprintf(bg.ac->str,"%.8lf",d);
  gtk_entry_set_text(GTK_ENTRY(bg.conststr),bg.ac->str);
  RedoRPNdecode(NULL,NULL);

  return(1);
}


void Set_fpos(int i)
{
  int b;

  for (b=0;b<bg.obj_num;b++) {
    body[b].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+2]=(double)i;
  }
}
void RPN_Dialog(char *s)
{
  GtkWidget *dialog,*label,*ok;

  dialog=gtk_dialog_new();
  label=gtk_label_new(s);
  ok=gtk_button_new_with_label("Ok");
  gtk_signal_connect_object (GTK_OBJECT (ok), "clicked",
		    GTK_SIGNAL_FUNC(gtk_widget_destroy), dialog);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),ok);
  gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),label);
  gtk_widget_show_all(dialog);
}

/* Set Constraint according to value */
void SetConstraintNum(GtkWidget *widget,gpointer data)
{
  int a;
  char s[24];

  a=(int)data;
  bg.ac=bg.param[bg.active_obj].constr;
  while ((bg.ac!=NULL) && (a)) {
    bg.ac=bg.ac->next;
    a--;
  }
  if (bg.ac==NULL) {
    bg.ac=bg.param[bg.active_obj].newconstr;
    gtk_widget_set_sensitive(bg.constdel,FALSE);
    gtk_widget_set_sensitive(bg.constadd,TRUE);
  } else {
    gtk_widget_set_sensitive(bg.constdel,TRUE);
    gtk_widget_set_sensitive(bg.constadd,FALSE);
  }
  gtk_option_menu_set_history(GTK_OPTION_MENU(bg.constpar),bg.ac->param/5);
  gtk_label_set_text(GTK_LABEL(bg.constparlabel),
  			cef_parstr[bg.ac->param/5][bg.ac->param%5]);
  gtk_option_menu_set_history(GTK_OPTION_MENU(bg.constrel),bg.ac->rel);
  sprintf(s,"%9.9lf",bg.ac->val);
  gtk_entry_set_text(GTK_ENTRY(bg.constval), s);
  if (bg.ac->str!=NULL) {
    gtk_entry_set_text(GTK_ENTRY(bg.conststr), bg.ac->str);
    gtk_option_menu_set_history(GTK_OPTION_MENU(bg.conststrmenu),
				bg.ac->strrel);
  } else {
    gtk_entry_set_text(GTK_ENTRY(bg.conststr), "");
    gtk_option_menu_set_history(GTK_OPTION_MENU(bg.conststrmenu),0);
  }
}

/* given the parameters set up the contraint window */
void SetUpConstraintFrame(void)
{
  int a,mask;
  CONSTRAINT *ctmp;
  char s[24];

  /* destroy previous number menu */
  if (bg.constrnummenusize>0) {
    for (a=0;a<bg.constrnummenusize;a++) {
      gtk_widget_destroy(bg.constrnummenuitem[a]);
    }
    free(bg.constrnummenuitem);
  }
  ctmp=bg.param[bg.active_obj].constr;
  /* count the number of contraints we have already */
  bg.constrnummenusize=1;
  while (ctmp!=NULL) {
    bg.constrnummenusize++;ctmp=ctmp->next;
  }
  gtk_widget_destroy(bg.constnummenu);
  bg.constnummenu=gtk_menu_new();
  bg.constrnummenuitem=(GtkWidget **)calloc(bg.constrnummenusize,
  			sizeof(GtkWidget *));
  for (a=0;a<bg.constrnummenusize-1;a++) {
    sprintf(s,"%d",a+1);
    bg.constrnummenuitem[a]=gtk_menu_item_new_with_label(s);
    gtk_signal_connect(GTK_OBJECT(bg.constrnummenuitem[a]),"activate",
		GTK_SIGNAL_FUNC(SetConstraintNum),(gpointer) a);
    gtk_widget_show(bg.constrnummenuitem[a]);
    gtk_menu_append(GTK_MENU(bg.constnummenu),bg.constrnummenuitem[a]);
  }
  bg.constrnummenuitem[a]=gtk_menu_item_new_with_label("New");
  gtk_signal_connect(GTK_OBJECT(bg.constrnummenuitem[a]),"activate",
		GTK_SIGNAL_FUNC(SetConstraintNum),(gpointer) a);
  gtk_widget_show(bg.constrnummenuitem[a]);
  gtk_menu_append(GTK_MENU(bg.constnummenu),bg.constrnummenuitem[a]);
  gtk_option_menu_set_menu (GTK_OPTION_MENU (bg.constnum),bg.constnummenu);
  a=max(bg.constrnummenusize-2,0);
  SetConstraintNum(NULL,(gpointer) a);
  gtk_option_menu_set_history(GTK_OPTION_MENU(bg.constnum),a);
  mask=BP_Pos;
  for (a=2;a<CEF_PARNUM;a++) {
    if (bg.param[bg.active_obj].par&mask) {
      gtk_widget_show(bg.parmainitem[a]);
    } else {
      gtk_widget_hide(bg.parmainitem[a]);
    }
    mask<<=1;
  }
  if (bg.param[bg.active_obj].type==BT_BallCC) {
    gtk_widget_show(bg.parmainitem[4]);
  }
}

/* Change blending */
void ShowConstraintFrame(GtkWidget *widget,gpointer data)
{
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))^1) {
    gtk_widget_show(bg.constframe);
  } else {
    gtk_widget_hide(bg.constframe);
  }
}

/* Set Contraint value */
void SetConstrValue(GtkWidget *widget,gpointer data)
{
  bg.ac->val=atof(gtk_entry_get_text(GTK_ENTRY(widget)));
}

/* Set Contraint relation */
void SetConstrRelation(GtkWidget *widget,gpointer data)
{
  bg.ac->rel=(int)data;
}

/* Set Contraint relation for string */
void SetStrConstrRelation(GtkWidget *widget,gpointer data)
{
  bg.ac->strrel=(int)data;
}

/* Change blending */
void BlendChange(GtkWidget *widget,gpointer data)
{
  bg.param[bg.active_obj].blend=gtk_toggle_button_get_active(
  			GTK_TOGGLE_BUTTON(widget))^1;
}

/* Change force scale value */
void Scale_Change(GtkWidget *widget,gpointer data)
{
  if (bg.param[bg.active_obj].scale_force) {
    bg.param[bg.active_obj].fsize=
  		atof(gtk_entry_get_text(GTK_ENTRY(widget)));
  } else {
    bg.param[bg.active_obj].fscale=
  		atof(gtk_entry_get_text(GTK_ENTRY(widget)));
  }
}

/* Change color value */
void Obj_Change_Color(GtkWidget *widget,gpointer data)
{
  bg.param[bg.active_obj].color[(int)data]=
  		atof(gtk_entry_get_text(GTK_ENTRY(widget)));
  RedrawTex();
}

/* Select texture or all */
void SetTextureType(GtkWidget *widget,gpointer data)
{
  int a;
  static GLfloat deflight[4][4] = {{ 0.0, 2.0, 3.0, 0.0},
    				    {1.0, 1.0, 1.0, 1.0},
				    {1.0, 1.0, 1.0, 1.0},
				    {0.0, 0.0, 0.0, 0.0}};
  GdkGLContext *glcontext = gtk_widget_get_gl_context (bg.texarea);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (bg.texarea);

  a=(int)data;
  bg.param[bg.active_obj].texture=a;
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext)) return;

  /* redraw glwindow */
  glLightfv(GL_LIGHT0,GL_POSITION,deflight[0]);
  glLightfv(GL_LIGHT0,GL_AMBIENT,deflight[1]);
  glLightfv(GL_LIGHT0,GL_DIFFUSE,deflight[2]);
  glLightfv(GL_LIGHT0,GL_SPECULAR,deflight[3]);
  glEnable(GL_LIGHTING);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_LIGHT0);
  glClearColor(gl.bg[0],gl.bg[1],gl.bg[2],gl.bg[3]);
  glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  if (a==0) {
    glColor4f( bg.param[bg.active_obj].color[0],
	bg.param[bg.active_obj].color[1],bg.param[bg.active_obj].color[2],
	bg.param[bg.active_obj].color[3]);
    glDisable(GL_TEXTURE_2D);
  } else {
    glColor4f(1.0,1.0,1.0,1.0);
    glBindTexture(GL_TEXTURE_2D,gl.texture.texture[a-1]);
    glEnable(GL_TEXTURE_2D);
  }
  glLoadIdentity();
  glTranslatef(-2.5,0.0,-4.0);
  glBegin(GL_QUADS);
  glNormal3f(0.0,0.0,1.0);
  glTexCoord2f(0.0, 1.0);
  glVertex3f(-1.5f, 1.5f, 0.0f);
  glTexCoord2f(1.0, 1.0);
  glVertex3f( 1.5f, 1.5f, 0.0f);
  glTexCoord2f(1.0, 0.0);
  glVertex3f( 1.5f,-1.5f, 0.0f);
  glTexCoord2f(0.0, 0.0);
  glVertex3f(-1.5f,-1.5f, 0.0f);
  glEnd();
  glTranslatef(4.0,0.0,0.0);
  gluSphere(gl.quad.q,1.5,gl.quad.sl,gl.quad.st);

  gdk_gl_drawable_swap_buffers (gldrawable);
  gdk_gl_drawable_gl_end (gldrawable);
}

/* Redraw texture gl window */
void RedrawTex()
{
  if (bg.obj_num==0) return;
  SetTextureType((GtkWidget *)NULL,(gpointer)
  		bg.param[bg.active_obj].texture);
}

/* switch force rescale */
void RescaleToMax(GtkWidget *widget)
{
  char s[24];

  bg.param[bg.active_obj].scale_force^=1;
  if (bg.param[bg.active_obj].scale_force) {
    sprintf(s,"%9.9lf",bg.param[bg.active_obj].fsize);
    gtk_label_set_text(GTK_LABEL(bg.fscalelabel),"Size of largest object:");
  } else {
    sprintf(s,"%9.9lf",bg.param[bg.active_obj].fscale);
    gtk_label_set_text(GTK_LABEL(bg.fscalelabel),"Multiply value by:");
  }
  gtk_entry_set_text(GTK_ENTRY(bg.fscalevalue), s);
}

/* show/do not show object */
void Show_Onoff(GtkWidget *widget,gpointer data)
{
  bg.param[(int)data].show^=1;
}

/* An object type is pressed so change subframe accordingly */
void Change_Subframe(GtkWidget *widget,gpointer data)
{
  int a;
  char s[24];

  a=(int)data;
  if (bg.param[a].type!=BT_XYZBall) {
    gtk_frame_set_label(GTK_FRAME(bg.frame),obj_name[bg.param[a].type]);
  } else {
    gtk_frame_set_label(GTK_FRAME(bg.frame),
    			XYZ_Known_Atoms_s[bg.param[a].atom]);
  }
  gtk_widget_show(bg.frame);
  bg.active_obj=a;
  gtk_option_menu_set_history(
  	GTK_OPTION_MENU (bg.textmenu),bg.param[a].texture);
  SetTextureType((GtkWidget *)NULL,(gpointer) bg.param[a].texture);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bg.blend),bg.param[a].blend);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[0]);
  gtk_entry_set_text(GTK_ENTRY(bg.r), s);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[1]);
  gtk_entry_set_text(GTK_ENTRY(bg.g), s);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[2]);
  gtk_entry_set_text(GTK_ENTRY(bg.b), s);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[3]);
  gtk_entry_set_text(GTK_ENTRY(bg.a), s);
  if (bg.param[bg.active_obj].scale_force) {
    sprintf(s,"%9.9lf",bg.param[bg.active_obj].fsize);
    gtk_label_set_text(GTK_LABEL(bg.fscalelabel),"Size of largest object:");
  } else {
    sprintf(s,"%9.9lf",bg.param[bg.active_obj].fscale);
    gtk_label_set_text(GTK_LABEL(bg.fscalelabel),"Multiply value by:");
  }
  gtk_entry_set_text(GTK_ENTRY(bg.fscalevalue), s);
  if (bg.constnummenu!=NULL) SetUpConstraintFrame();
}

/* set label and type */
void SetParameterType(GtkWidget *widget,gpointer data)
{
  int a;

  a=(int) data;
  gtk_label_set_text(GTK_LABEL(bg.constparlabel),cef_parstr[a/5][a%5]);
  bg.ac->param=a;
}

/* Change between rad/deg */
void DegRad(GtkWidget *widget, gpointer data)
{
  GtkWidget *dialog,*label,*ok;
  const char *dr[2]= {	"Radian unit is selected",
    			"Degrees unit is selected" };

  RPN_deg^=1;
  dialog=gtk_dialog_new();
  label=gtk_label_new(dr[RPN_deg]);
  ok=gtk_button_new_with_label("Ok");
  gtk_signal_connect_object (GTK_OBJECT (ok), "clicked",
		    GTK_SIGNAL_FUNC(gtk_widget_destroy), dialog);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),ok);
  gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),label);
  gtk_widget_show_all(dialog);
}

/* constraint string is changed redo RPN encoding */
void RedoRPNdecode(GtkWidget *widget,gpointer data)
{
  int dofree;
  RPN *pr;
  char *error;

  if (bg.ac!=bg.param[bg.active_obj].newconstr) {
    if (bg.ac->str!=NULL) {
      free(bg.ac->str);
      dofree=1;
    } else dofree=0;
    bg.ac->str=strdup(gtk_entry_get_text(GTK_ENTRY(bg.conststr)));
    if (strlen(bg.ac->str)) {
      if (!RPN_decode(bg.ac->str,&pr,&error,RpnVT_Constr)) {
	RPN_Dialog(error);
	free(error);
	return;
      }
      if (dofree) RPN_Free(&bg.ac->rpnroot);
      bg.ac->rpnroot=pr;
    } else {
      free(bg.ac->str);
      if (dofree) RPN_Free(&bg.ac->rpnroot);
      bg.ac->str=NULL;
    }
  }
}

/* Constraint accepted */
void Constraint_add(GtkWidget *widget,gpointer data)
{
  char *error,*s;
  RPN *pr;
  CONSTRAINT *ctmp;

  ctmp=bg.param[bg.active_obj].constr;
  if (strlen(gtk_entry_get_text(GTK_ENTRY(bg.conststr)))) {
    s=strdup(gtk_entry_get_text(GTK_ENTRY(bg.conststr)));
    if (!RPN_decode(s,&pr,&error,RpnVT_Constr)) {
      RPN_Dialog(error);
      free(s);
      free(error);
      return;
    }
  }
  if (ctmp==NULL) {
    ctmp=bg.param[bg.active_obj].constr=
    			(CONSTRAINT *)malloc(sizeof(CONSTRAINT));
  } else {
    while (ctmp->next!=NULL) ctmp=ctmp->next;
    ctmp->next=(CONSTRAINT *)malloc(sizeof(CONSTRAINT));
    ctmp=ctmp->next;
  }
  if (strlen(gtk_entry_get_text(GTK_ENTRY(bg.conststr)))) {
    ctmp->str=s;
    ctmp->rpnroot=pr;
  } else {
    ctmp->str=NULL;
  }
  ctmp->strrel=bg.ac->strrel;
  ctmp->param=bg.ac->param;
  ctmp->rel=bg.ac->rel;
  ctmp->val=bg.ac->val;
  ctmp->next=NULL;
  bg.param[bg.active_obj].constr_num++;
  bg.ac->param=0;
  bg.ac->rel=1;
  bg.ac->strrel=0;
  bg.ac->val= -1.0;
  bg.ac->str=NULL;
  bg.ac->next=NULL;
  bg.ac=ctmp;
  SetUpConstraintFrame();
}

/* Constraint deleted */
void Constraint_del(GtkWidget *widget,gpointer data)
{
  CONSTRAINT *ctmp,*ctmp2;

  if (bg.ac->str!=NULL) {
    free(bg.ac->str);
    RPN_Free(&bg.ac->rpnroot);
  }
  ctmp=bg.param[bg.active_obj].constr;
  if (bg.ac==ctmp) {
    bg.param[bg.active_obj].constr=ctmp->next;
    free(ctmp);
  } else {
    while ((ctmp->next!=NULL) && (ctmp->next!=bg.ac)) ctmp=ctmp->next;
    if (ctmp->next==NULL) return;
    ctmp2=ctmp->next;
    ctmp->next=ctmp2->next;
    free(ctmp2);
  }
  bg.param[bg.active_obj].constr_num--;
  SetUpConstraintFrame();
}

void CreateObjectFrame(GtkWidget *frame,GdkGLConfig *glconfig,
					GtkAccelGroup *accel_group)
{
  GtkWidget *vbox;		/* vbox */
  GtkWidget **hbox;		/* hboxes */
  GtkWidget **button;		/* Buttons */
  GtkWidget **oonoff;		/* chkbutton for objects */
  GtkWidget *off;		/* switch frame off button */
  GtkWidget *fvbox;		/* vbox for frame */
  GtkWidget *hbox1;		/* hbox1 for frame */
  GtkWidget *hbox2;		/* hbox2 for frame */
  GtkWidget *hbox3;		/* hbox3 for frame */
  GtkWidget *hbox4;		/* hbox4 for frame */
  GtkWidget *svbox1;		/* vbox for hbox2 frame */
  GtkWidget *svbox2;		/* vbox for hbox2 frame */
  GtkWidget *shbox1;		/* hbox for svbox1 */
  GtkWidget *shbox2;		/* hbox for svbox1 */
  GtkWidget *shbox3;		/* hbox for svbox2 */
  GtkWidget *shbox4;		/* hbox for svbox2 */
  GtkWidget *label;		/* label widget */
  GtkWidget *menu;		/* menu widget of latter */
  GtkWidget *item;		/* item for the latter */
  GtkWidget *foff;		/* switch off sub frame */
  GtkWidget *cvbox;		/* constframe vbox */
  GtkWidget *chbox1;		/* constframe hbox */
  GtkWidget *chbox2;		/* constframe hbox */
  GtkWidget *constrelmenu;	/* menu widget of latter */
  GtkWidget *constparmenu;	/* constraints parameter menu */
  GtkWidget *conststrmmenu;	/* constraints parameter menu for string */
  int i,j;
  char s[24];

  if (bg.obj_num==0) return;
  RPN_deg=1;
  bg.constnummenu=NULL;
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(frame),vbox);
  hbox=(GtkWidget **) calloc(bg.obj_num/3+1,sizeof(GtkWidget *));
  button=(GtkWidget **) calloc(bg.obj_num,sizeof(GtkWidget *));
  oonoff=(GtkWidget **) calloc(bg.obj_num,sizeof(GtkWidget *));
  for (i=0;i<bg.obj_num/3+1;i++) {
    hbox[i] = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox),hbox[i],TRUE,TRUE,3);
    gtk_widget_show(hbox[i]);
  }
  for (i=0;i<bg.obj_num;i++) {
    if (bg.param[i].type!=BT_XYZBall) {
      button[i]=gtk_button_new_with_label(obj_name[bg.param[i].type]);
    } else {
      button[i]=gtk_button_new_with_label(XYZ_Known_Atoms_s[bg.param[i].atom]);
    }
    gtk_box_pack_start(GTK_BOX(hbox[i/3]),button[i],FALSE,FALSE,5);
    gtk_signal_connect(GTK_OBJECT (button[i]),"clicked",
            GTK_SIGNAL_FUNC(Change_Subframe),(gpointer)i);
    gtk_widget_show(button[i]);
    oonoff[i]=gtk_check_button_new();
    gtk_box_pack_start(GTK_BOX(hbox[i/3]),oonoff[i],FALSE,FALSE,5);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(oonoff[i]), TRUE);
    gtk_signal_connect(GTK_OBJECT (oonoff[i]),"clicked",
            GTK_SIGNAL_FUNC(Show_Onoff),(gpointer)i);
    gtk_widget_show(oonoff[i]);
  }
  off=gtk_check_button_new_with_label("Off");
  gtk_box_pack_end(GTK_BOX(hbox[(bg.obj_num-1)/3]),off,FALSE,FALSE,0);
  gtk_widget_show(off);

  /* inner frame */
  bg.active_obj=0;
  if (bg.param[bg.active_obj].type!=BT_XYZBall) {
    bg.frame=gtk_frame_new(obj_name[bg.param[bg.active_obj].type]);
  } else {
    bg.frame=gtk_frame_new(XYZ_Known_Atoms_s[bg.param[bg.active_obj].atom]);
  }
  gtk_box_pack_start(GTK_BOX(vbox),bg.frame,TRUE,TRUE,3);
  fvbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(bg.frame),fvbox);
  hbox1 = gtk_hbox_new(FALSE, 0);
  hbox2 = gtk_hbox_new(FALSE, 0);
  bg.forcehbox=hbox3 = gtk_hbox_new(FALSE, 0);
  hbox4 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(fvbox),hbox1,TRUE,TRUE,3);
  gtk_box_pack_start(GTK_BOX(fvbox),hbox2,TRUE,TRUE,3);
  gtk_box_pack_start(GTK_BOX(fvbox),hbox3,TRUE,TRUE,3);
  gtk_box_pack_start(GTK_BOX(fvbox),hbox4,TRUE,TRUE,3);
  label=gtk_label_new("Texture:");
  gtk_box_pack_start(GTK_BOX(hbox1),label,FALSE,FALSE,5);
  gtk_widget_show(label);
  bg.textmenu=gtk_option_menu_new();
  menu=gtk_menu_new();
  for (i=0;i<=gl.texture.num;i++) {
    if (i>0) sprintf(s,"%d",i-1);
    else sprintf(s,"No");
    item=gtk_menu_item_new_with_label(s);
    gtk_signal_connect(GTK_OBJECT(item),"activate",
	GTK_SIGNAL_FUNC(SetTextureType),(gpointer) i);
    gtk_widget_show(item);
    gtk_menu_append (GTK_MENU (menu), item);
  }
  gtk_option_menu_set_menu (GTK_OPTION_MENU (bg.textmenu), menu);
  gtk_widget_show(menu);
  gtk_widget_show(bg.textmenu);
  gtk_box_pack_start(GTK_BOX(hbox1),bg.textmenu,FALSE,FALSE,5);
  label=gtk_label_new("Color:");
  gtk_box_pack_end(GTK_BOX(hbox1),label,FALSE,FALSE,5);
  gtk_widget_show(label);

  bg.texarea = gtk_drawing_area_new();
  gtk_widget_set_size_request (bg.texarea, 150, 60);
  gtk_widget_set_usize(bg.texarea, 150, 60);
  gtk_widget_set_gl_capability(bg.texarea,glconfig,NULL,TRUE,GDK_GL_RGBA_TYPE);
  gtk_box_pack_start(GTK_BOX(hbox2),bg.texarea,FALSE,FALSE,10);
  gtk_widget_show(bg.texarea);

  svbox1 = gtk_vbox_new(FALSE, 0);
  svbox2 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_end(GTK_BOX(hbox2),svbox2,FALSE,FALSE,3);
  gtk_box_pack_end(GTK_BOX(hbox2),svbox1,FALSE,FALSE,3);
  shbox1=gtk_hbox_new(FALSE,0);
  shbox2=gtk_hbox_new(FALSE,0);
  shbox3=gtk_hbox_new(FALSE,0);
  shbox4=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(svbox1),shbox1,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(svbox1),shbox2,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(svbox2),shbox3,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(svbox2),shbox4,FALSE,FALSE,3);
  label=gtk_label_new("Red:");
  gtk_box_pack_start(GTK_BOX(shbox1),label,FALSE,FALSE,3);
  gtk_widget_show(label);
  bg.r=gtk_entry_new_with_max_length(4);
  gtk_widget_set_usize(bg.r,36,0);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[0]);
  gtk_entry_set_text(GTK_ENTRY(bg.r), s);
  gtk_box_pack_end(GTK_BOX(shbox1), bg.r, FALSE, FALSE, 3);
  gtk_widget_show(bg.r);
  label=gtk_label_new("Green:");
  gtk_box_pack_start(GTK_BOX(shbox3),label,FALSE,FALSE,3);
  gtk_widget_show(label);
  bg.g=gtk_entry_new_with_max_length(4);
  gtk_widget_set_usize(bg.g,36,0);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[1]);
  gtk_entry_set_text(GTK_ENTRY(bg.g), s);
  gtk_box_pack_end(GTK_BOX(shbox3), bg.g, FALSE, FALSE, 3);
  gtk_widget_show(bg.g);
  label=gtk_label_new("Blue:");
  gtk_box_pack_start(GTK_BOX(shbox2),label,FALSE,FALSE,3);
  gtk_widget_show(label);
  bg.b=gtk_entry_new_with_max_length(4);
  gtk_widget_set_usize(bg.b,36,0);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[2]);
  gtk_entry_set_text(GTK_ENTRY(bg.b), s);
  gtk_box_pack_end(GTK_BOX(shbox2), bg.b, FALSE, FALSE, 3);
  gtk_widget_show(bg.b);
  label=gtk_label_new("Alpha:");
  gtk_box_pack_start(GTK_BOX(shbox4),label,FALSE,FALSE,3);
  gtk_widget_show(label);
  bg.a=gtk_entry_new_with_max_length(4);
  gtk_widget_set_usize(bg.a,36,0);
  sprintf(s,"%.2lf",bg.param[bg.active_obj].color[3]);
  gtk_entry_set_text(GTK_ENTRY(bg.a), s);
  gtk_box_pack_end(GTK_BOX(shbox4), bg.a, FALSE, FALSE, 3);
  gtk_widget_show(bg.a);
  gtk_widget_show(shbox1);
  gtk_widget_show(shbox2);
  gtk_widget_show(shbox3);
  gtk_widget_show(shbox4);
  gtk_widget_show(svbox1);
  gtk_widget_show(svbox2);

  bg.forcescale=gtk_check_button_new_with_label("Rescale to max");
  gtk_box_pack_start(GTK_BOX(hbox3), bg.forcescale, FALSE, FALSE, 3);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bg.forcescale),
		 			bg.param[bg.active_obj].scale_force);
  gtk_widget_show(bg.forcescale);
  bg.fscalevalue=gtk_entry_new_with_max_length(20);
  gtk_widget_set_usize(bg.fscalevalue,60,0);
  gtk_box_pack_end(GTK_BOX(hbox3), bg.fscalevalue, FALSE, FALSE, 3);
  gtk_widget_show(bg.fscalevalue);
  if (bg.param[bg.active_obj].scale_force) {
    sprintf(s,"%9.9lf",bg.param[bg.active_obj].fsize);
    bg.fscalelabel=gtk_label_new("Size of largest object:");
  } else {
    sprintf(s,"%9.9lf",bg.param[bg.active_obj].fscale);
    bg.fscalelabel=gtk_label_new("Multiply value by:");
  }
  gtk_entry_set_text(GTK_ENTRY(bg.fscalevalue), s);
  gtk_box_pack_end(GTK_BOX(hbox3),bg.fscalelabel,FALSE,FALSE,3);
  gtk_widget_show(bg.fscalelabel);

  foff=gtk_check_button_new_with_label("Frame off");
  gtk_box_pack_end(GTK_BOX(hbox4),foff,FALSE,FALSE,0);
  gtk_widget_show(foff);
  bg.blend=gtk_check_button_new_with_label("Transparent");
  gtk_box_pack_start(GTK_BOX(hbox4), bg.blend, FALSE, FALSE, 3);
  gtk_widget_show(bg.blend);
  bg.constraintbutt=gtk_check_button_new_with_label("Constraints");
  gtk_box_pack_start(GTK_BOX(hbox4), bg.constraintbutt, FALSE, FALSE, 20);
  gtk_widget_show(bg.constraintbutt);

  Change_Subframe((GtkWidget *)NULL,(gpointer)bg.active_obj);

  /* Constraints frame */
  bg.constframe=gtk_frame_new("Constraints");
  gtk_box_pack_start(GTK_BOX(fvbox),bg.constframe,TRUE,TRUE,3);
  cvbox=gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(bg.constframe),cvbox);
  chbox1=gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cvbox),chbox1,FALSE,FALSE,3);
  chbox2=gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cvbox),chbox2,FALSE,FALSE,3);
  bg.constnum=gtk_option_menu_new();
  bg.constpar=gtk_option_menu_new();
  bg.constparlabel=gtk_label_new("num");
  constrelmenu=gtk_menu_new();
  bg.constrel=gtk_option_menu_new();
  bg.constval=gtk_entry_new_with_max_length(18);
  gtk_widget_set_usize(bg.constval,50,0);
  bg.constdel=gtk_button_new_with_label("Del");
  bg.constadd=gtk_button_new_with_label("Add");

  gtk_box_pack_start(GTK_BOX(chbox1),bg.constnum,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox1),bg.constpar,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox1),bg.constparlabel,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox1),bg.constrel,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox1),bg.constval,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox1),bg.constdel,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox1),bg.constadd,FALSE,FALSE,3);
  bg.constnummenu=gtk_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(bg.constnum),bg.constnummenu);
  bg.constrnummenuitem=NULL;
  bg.constrnummenusize=0;
  gtk_widget_show(bg.constnummenu);
  constparmenu=gtk_menu_new();
  for (i=0;i<CEF_PARNUM;i++) {
    bg.parmainitem[i]=gtk_menu_item_new_with_label(cef_parmainstr[i]);
    menu=gtk_menu_new();
    j=0;
    while ((j<CEF_SUBPARNUM) && (cef_parstr[i][j][0]!='\0')) {
      item=gtk_menu_item_new_with_label(cef_parstr[i][j]);
      gtk_signal_connect(GTK_OBJECT(item),"activate",
	      GTK_SIGNAL_FUNC(SetParameterType),(gpointer)
	      (i*CEF_SUBPARNUM+j));
      gtk_widget_show(item);
      gtk_menu_append (GTK_MENU(menu),item);
      j++;
    }
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(bg.parmainitem[i]),menu);
    gtk_widget_show(menu);
    gtk_widget_show(bg.parmainitem[i]);
    gtk_menu_append (GTK_MENU (constparmenu), bg.parmainitem[i]);
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(bg.constpar),constparmenu);
  gtk_widget_show(constparmenu);

  for (i=0;i<sizeof(cef_relstr)/2;i++) {
    item=gtk_menu_item_new_with_label(cef_relstr[i]);
    gtk_signal_connect(GTK_OBJECT(item),"activate",
	      GTK_SIGNAL_FUNC(SetConstrRelation),(gpointer) i);
    gtk_widget_show(item);
    gtk_menu_append (GTK_MENU (constrelmenu), item);
  }
  gtk_option_menu_set_menu (GTK_OPTION_MENU (bg.constrel),constrelmenu);
  gtk_widget_show(constrelmenu);

  conststrmmenu=gtk_menu_new();
  bg.conststrmenu=gtk_option_menu_new();
  for (i=0;i<sizeof(cef_strrelstr)/6;i++) {
    item=gtk_menu_item_new_with_label(cef_strrelstr[i]);
    gtk_signal_connect(GTK_OBJECT(item),"activate",
	      GTK_SIGNAL_FUNC(SetStrConstrRelation),(gpointer) i);
    gtk_widget_show(item);
    gtk_menu_append (GTK_MENU (conststrmmenu), item);
  }
  gtk_option_menu_set_menu (GTK_OPTION_MENU(bg.conststrmenu),conststrmmenu);
  gtk_widget_show(conststrmmenu);
  bg.conststr=gtk_entry_new_with_max_length(255);
  gtk_box_pack_start(GTK_BOX(chbox2),bg.conststrmenu,FALSE,FALSE,3);
  gtk_box_pack_start(GTK_BOX(chbox2),bg.conststr,TRUE,TRUE,3);

  gtk_widget_show(bg.constnum);
  gtk_widget_show(bg.constpar);
  gtk_widget_show(bg.constparlabel);
  gtk_widget_show(bg.constrel);
  gtk_widget_show(bg.constval);
  gtk_widget_show(bg.constdel);
  gtk_widget_show(bg.constadd);
  gtk_widget_show(bg.conststrmenu);
  gtk_widget_show(bg.conststr);
  gtk_widget_show(chbox1);
  gtk_widget_show(chbox2);
  gtk_widget_show(cvbox);

  gtk_widget_show(hbox4);
  gtk_widget_show(hbox3);
  gtk_widget_show(hbox2);
  gtk_widget_show(hbox1);
  gtk_widget_show(fvbox);
  gtk_widget_show(bg.frame);

  /* Set up gl box */
  {
    GdkGLContext *glcontext = gtk_widget_get_gl_context (bg.texarea);
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (bg.texarea);
    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
      glClearColor(gl.bg[0],gl.bg[1],gl.bg[2],gl.bg[3]);
      glClearDepth(1.0);
      glDepthFunc(GL_LESS);
      glEnable(GL_DEPTH_TEST);
      glShadeModel(GL_SMOOTH);
      glViewport(0,0,150,60);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(gl.persp.angle,2.5,gl.persp.zNear,gl.persp.zFar);
      glMatrixMode(GL_MODELVIEW);

      for (i=0;i<gl.texture.num;i++) {
	glBindTexture(GL_TEXTURE_2D, gl.texture.texture[i]);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, gl.texture.detail[i].width,
			gl.texture.detail[i].height, 0,
			GL_RGB, GL_UNSIGNED_BYTE, gl.texture.detail[i].data);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
      }
      gdk_gl_drawable_gl_end (gldrawable);
    }
  }

  gtk_signal_connect (GTK_OBJECT(bg.conststr), "activate",
  		GTK_SIGNAL_FUNC(RedoRPNdecode),NULL);
  gtk_signal_connect (GTK_OBJECT(bg.constadd), "clicked",
  		GTK_SIGNAL_FUNC(Constraint_add),NULL);
  gtk_signal_connect (GTK_OBJECT(bg.constdel), "clicked",
  		GTK_SIGNAL_FUNC(Constraint_del),NULL);
  gtk_signal_connect (GTK_OBJECT (bg.r), "changed",
     		GTK_SIGNAL_FUNC(Obj_Change_Color),(gpointer)0);
  gtk_signal_connect (GTK_OBJECT (bg.g), "changed",
     		GTK_SIGNAL_FUNC(Obj_Change_Color),(gpointer)1);
  gtk_signal_connect (GTK_OBJECT (bg.b), "changed",
     		GTK_SIGNAL_FUNC(Obj_Change_Color),(gpointer)2);
  gtk_signal_connect (GTK_OBJECT (bg.a), "changed",
     		GTK_SIGNAL_FUNC(Obj_Change_Color),(gpointer)3);
  gtk_signal_connect (GTK_OBJECT (off), "toggled",
		GTK_SIGNAL_FUNC (FrameOff), (gpointer)frame);
  gtk_signal_connect (GTK_OBJECT (foff), "toggled",
		GTK_SIGNAL_FUNC (FrameOff), (gpointer)bg.frame);
  gtk_signal_connect(GTK_OBJECT(bg.texarea), "expose_event",
		GTK_SIGNAL_FUNC(RedrawTex), NULL);
  gtk_signal_connect (GTK_OBJECT (bg.blend), "pressed",
		GTK_SIGNAL_FUNC (BlendChange), NULL);
  gtk_signal_connect (GTK_OBJECT (bg.forcescale), "pressed",
		GTK_SIGNAL_FUNC (RescaleToMax), NULL);
  gtk_signal_connect (GTK_OBJECT (bg.fscalevalue), "activate",
     		GTK_SIGNAL_FUNC(Scale_Change),NULL);
  gtk_signal_connect (GTK_OBJECT (bg.constraintbutt), "pressed",
		GTK_SIGNAL_FUNC (ShowConstraintFrame), NULL);
  gtk_signal_connect (GTK_OBJECT (bg.constval), "changed",
		GTK_SIGNAL_FUNC (SetConstrValue), NULL);

  gtk_widget_show(vbox);
  SetUpConstraintFrame();
}
